@@ -1,16 +1,16 @@
-## Main function: BDMCMC algorithm for graphical models 
-################################################################################
+## ----------------------------------------------------------------------------|
+# Main function of BDgraph package: BDMCMC algorithm for graphical models 
+## ----------------------------------------------------------------------------|
 bdgraph = function( data, n = NULL, method = "ggm", algorithm = "bdmcmc", 
 					iter = 5000, burnin = iter / 2, g.start = "empty", prior.df = 3,
-					multi.update = NULL, save.all = FALSE )
+					multi.update = NULL, save.all = FALSE, print = 1000 )
 {
-	threshold = 1e-8  # for sampling from gwishart distribution
 	burnin    = floor( burnin )
 	
-	if( class(data) == "sim" ) data <- data $ data
+	if( class( data ) == "sim" ) data <- data $ data
 
-	if( !is.matrix(data) & !is.data.frame(data) ) stop( "Data should be a matrix or dataframe" )
-	if( is.data.frame(data) ) data <- data.matrix(data)
+	if( !is.matrix( data ) & !is.data.frame( data ) ) stop( "Data should be a matrix or dataframe" )
+	if( is.data.frame( data ) ) data <- data.matrix( data )
 	if( iter <= burnin )   stop( "Number of iteration must be more than number of burn-in" )
 
 	if( any( is.na( data ) ) ) 
@@ -21,9 +21,9 @@ bdgraph = function( data, n = NULL, method = "ggm", algorithm = "bdmcmc",
 		gcgm_NA = 0
 	}
 		
-	dimd <- dim(data)
+	dimd <- dim( data )
 	p    <- dimd[2]
-	if( is.null(n) ) n <- dimd[1]
+	if( is.null( n ) ) n <- dimd[1]
 
 	if( method == "gcgm" )
 	{
@@ -62,7 +62,7 @@ bdgraph = function( data, n = NULL, method = "ggm", algorithm = "bdmcmc",
 		}
 	}
    
-	D      = diag(p)
+	D      = diag( p )
 	b      = prior.df
 	b_star = b + n
 	Ds     = D + S
@@ -83,20 +83,20 @@ bdgraph = function( data, n = NULL, method = "ggm", algorithm = "bdmcmc",
 	
 	if( class( g.start ) == "character" && g.start == "empty"  )
 	{
-		G = 0 * S
+		G = matrix( 0, p, p )
 		K = G
 		
-		result = .C( "rgwish", as.integer(G), as.double(Ts), K = as.double(K), as.integer(b_star), as.integer(p), as.double(threshold), PACKAGE = "BDgraph" )
+		result = .C( "rgwish_c", as.integer(G), as.double(Ts), K = as.double(K), as.integer(b_star), as.integer(p), PACKAGE = "BDgraph" )
 		K      = matrix ( result $ K, p, p ) 
 	}
 	
 	if( class( g.start ) == "character" && g.start == "full" )
 	{
-		G       = matrix(1, p, p)
-		diag(G) = 0
-		K       = 0 * G
+		G         = matrix( 1, p, p )
+		diag( G ) = 0
+		K         = matrix( 0, p, p )
 
-		result = .C( "rwish", as.double(Ts), K = as.double(K), as.integer(b_star), as.integer(p), PACKAGE = "BDgraph" )
+		result = .C( "rwish_c", as.double(Ts), K = as.double(K), as.integer(b_star), as.integer(p), PACKAGE = "BDgraph" )
 		K      = matrix ( result $ K, p, p ) 
 	}	
 
@@ -105,8 +105,8 @@ bdgraph = function( data, n = NULL, method = "ggm", algorithm = "bdmcmc",
 		G       = g.start
 		diag(G) = 0
 		
-		K       = 0 * G
-		result = .C( "rgwish", as.integer(G), as.double(Ts), K = as.double(K), as.integer(b_star), as.integer(p), as.double(threshold), PACKAGE = "BDgraph" )
+		K      = matrix( 0, p, p )
+		result = .C( "rgwish_c", as.integer(G), as.double(Ts), K = as.double(K), as.integer(b_star), as.integer(p), PACKAGE = "BDgraph" )
 		K      = matrix ( result $ K, p, p ) 	
 	}
 			
@@ -120,7 +120,7 @@ bdgraph = function( data, n = NULL, method = "ggm", algorithm = "bdmcmc",
 		all_weights   = c( rep ( 1, iter - burnin ) )         # waiting time for every state		
 		size_sample_g = 0
 	}else{
-		p_links = 0 * K
+		p_links = matrix( 0, p, p )
 	}
 
 	if( ( save.all == TRUE ) && ( p > 50 & iter > 20000 ) )
@@ -129,17 +129,18 @@ bdgraph = function( data, n = NULL, method = "ggm", algorithm = "bdmcmc",
 		print( ( iter - burnin ) * object.size( string_g ), units = "auto" ) 
 	} 
 	
-	K_hat      = 0 * K
+	K_hat      = matrix( 0, p, p )
 	last_graph = K_hat
 	last_K     = K_hat
 
 	if( ( is.null( multi.update ) ) && ( p > 10 & iter > ( 5000 / p ) ) )
-	{
 		multi.update = floor( p / 10 )
-	}
 	
 	if( is.null( multi.update ) ) multi.update = 1
 	multi_update = multi.update
+
+	if( ( p < 10 ) && ( multi_update > 1 ) )      cat( " WARNING: for the cases p < 10, multi.update must be 1 " )
+	if( multi_update > min( p, sqrt( p * 11 ) ) ) cat( " WARNING: multi.update must be smaller " )
 	
 	mes <- paste( c( iter, " iteration is started.                    " ), collapse = "" )
 	cat( mes, "\r" )
@@ -151,15 +152,16 @@ bdgraph = function( data, n = NULL, method = "ggm", algorithm = "bdmcmc",
 			result = .C( "ggm_bdmcmc_map", as.integer(iter), as.integer(burnin), G = as.integer(G), as.double(Ts), K = as.double(K), as.integer(p), 
 						all_graphs = as.integer(all_graphs), all_weights = as.double(all_weights), K_hat = as.double(K_hat), 
 						sample_graphs = as.character(sample_graphs), graph_weights = as.double(graph_weights), size_sample_g = as.integer(size_sample_g),
-						as.integer(b), as.integer(b_star), as.double(Ds), as.double(threshold), PACKAGE = "BDgraph" )
+						as.integer(b), as.integer(b_star), as.double(Ds), as.integer(print), PACKAGE = "BDgraph" )
 		}
 
 		if( ( method == "ggm" ) && ( algorithm == "bdmcmc" ) && ( multi_update != 1 ) )
 		{
+			counter_all_g = 0
 			result = .C( "ggm_bdmcmc_map_multi_update", as.integer(iter), as.integer(burnin), G = as.integer(G), as.double(Ts), K = as.double(K), as.integer(p), 
 						all_graphs = as.integer(all_graphs), all_weights = as.double(all_weights), K_hat = as.double(K_hat), 
-						sample_graphs = as.character(sample_graphs), graph_weights = as.double(graph_weights), size_sample_g = as.integer(size_sample_g),
-						as.integer(b), as.integer(b_star), as.double(Ds), as.double(threshold), as.integer(multi_update), PACKAGE = "BDgraph" )
+						sample_graphs = as.character(sample_graphs), graph_weights = as.double(graph_weights), size_sample_g = as.integer(size_sample_g), counter_all_g = as.integer(counter_all_g),
+						as.integer(b), as.integer(b_star), as.double(Ds), as.integer(multi_update), as.integer(print), PACKAGE = "BDgraph" )
 		}
 
 		if( ( method == "ggm" ) && ( algorithm == "rjmcmc" ) )
@@ -167,7 +169,7 @@ bdgraph = function( data, n = NULL, method = "ggm", algorithm = "bdmcmc",
 			result = .C( "ggm_rjmcmc_map", as.integer(iter), as.integer(burnin), G = as.integer(G), as.double(Ts), K = as.double(K), as.integer(p), 
 						all_graphs = as.integer(all_graphs), all_weights = as.double(all_weights), K_hat = as.double(K_hat), 
 						sample_graphs = as.character(sample_graphs), graph_weights = as.double(graph_weights), size_sample_g = as.integer(size_sample_g),
-						as.integer(b), as.integer(b_star), as.double(Ds), as.double(threshold), PACKAGE = "BDgraph" )
+						as.integer(b), as.integer(b_star), as.double(Ds), as.integer(print), PACKAGE = "BDgraph" )
 		}
 		
 		if( ( method == "gcgm" ) && ( algorithm == "bdmcmc" ) && ( multi_update == 1 ) )
@@ -176,16 +178,17 @@ bdgraph = function( data, n = NULL, method = "ggm", algorithm = "bdmcmc",
 						as.double(Z), as.integer(R), as.integer(n), as.integer(gcgm_NA),
 						all_graphs = as.integer(all_graphs), all_weights = as.double(all_weights), K_hat = as.double(K_hat), 
 						sample_graphs = as.character(sample_graphs), graph_weights = as.double(graph_weights), size_sample_g = as.integer(size_sample_g),
-						as.integer(b), as.integer(b_star), as.double(D), as.double(Ds), as.double(threshold), PACKAGE = "BDgraph" )
+						as.integer(b), as.integer(b_star), as.double(D), as.double(Ds), as.integer(print), PACKAGE = "BDgraph" )
 		}
 
 		if( ( method == "gcgm" ) && ( algorithm == "bdmcmc" ) && ( multi_update != 1 ) )
 		{
+			counter_all_g = 0
 			result = .C( "gcgm_bdmcmc_map_multi_update", as.integer(iter), as.integer(burnin), G = as.integer(G), as.double(Ts), K = as.double(K), as.integer(p),
 						as.double(Z), as.integer(R), as.integer(n), as.integer(gcgm_NA),
 						all_graphs = as.integer(all_graphs), all_weights = as.double(all_weights), K_hat = as.double(K_hat), 
-						sample_graphs = as.character(sample_graphs), graph_weights = as.double(graph_weights), size_sample_g = as.integer(size_sample_g),
-						as.integer(b), as.integer(b_star), as.double(D), as.double(Ds), as.double(threshold), as.integer(multi_update), PACKAGE = "BDgraph" )
+						sample_graphs = as.character(sample_graphs), graph_weights = as.double(graph_weights), size_sample_g = as.integer(size_sample_g), counter_all_g = as.integer(counter_all_g),
+						as.integer(b), as.integer(b_star), as.double(D), as.double(Ds), as.integer(multi_update), as.integer(print), PACKAGE = "BDgraph" )
 		}
    
 		if( ( method == "gcgm" ) && ( algorithm == "rjmcmc" ) )
@@ -194,59 +197,61 @@ bdgraph = function( data, n = NULL, method = "ggm", algorithm = "bdmcmc",
 						as.double(Z), as.integer(R), as.integer(n), as.integer(gcgm_NA),
 						all_graphs = as.integer(all_graphs), all_weights = as.double(all_weights), K_hat = as.double(K_hat), 
 						sample_graphs = as.character(sample_graphs), graph_weights = as.double(graph_weights), size_sample_g = as.integer(size_sample_g),
-						as.integer(b), as.integer(b_star), as.double(D), as.double(Ds), as.double(threshold), PACKAGE = "BDgraph" )
+						as.integer(b), as.integer(b_star), as.double(D), as.double(Ds), as.integer(print), PACKAGE = "BDgraph" )
 		}	
 
 		# for Double Metropolis-Hasting 
-		if( ( method == "ggm" ) && ( algorithm == "bd_dmh" ) && ( multi_update == 1 ) )
+		if( ( method == "ggm" ) && ( algorithm == "bd-dmh" ) && ( multi_update == 1 ) )
 		{
 			result = .C( "ggm_DMH_bdmcmc_map", as.integer(iter), as.integer(burnin), G = as.integer(G), as.double(Ts), as.double(Ti), K = as.double(K), as.integer(p), 
 						all_graphs = as.integer(all_graphs), all_weights = as.double(all_weights), K_hat = as.double(K_hat), 
 						sample_graphs = as.character(sample_graphs), graph_weights = as.double(graph_weights), size_sample_g = as.integer(size_sample_g),
-						as.integer(b), as.integer(b_star), as.double(Ds), as.double(D), as.double(threshold), PACKAGE = "BDgraph" )
+						as.integer(b), as.integer(b_star), as.double(Ds), as.double(D), as.integer(print), PACKAGE = "BDgraph" )
 		}
 
-		if( ( method == "ggm" ) && ( algorithm == "bd_dmh" ) && ( multi_update != 1 ) )
+		if( ( method == "ggm" ) && ( algorithm == "bd-dmh" ) && ( multi_update != 1 ) )
 		{
+			counter_all_g = 0
 			result = .C( "ggm_DMH_bdmcmc_map_multi_update", as.integer(iter), as.integer(burnin), G = as.integer(G), as.double(Ts), as.double(Ti), K = as.double(K), as.integer(p), 
 						all_graphs = as.integer(all_graphs), all_weights = as.double(all_weights), K_hat = as.double(K_hat), 
-						sample_graphs = as.character(sample_graphs), graph_weights = as.double(graph_weights), size_sample_g = as.integer(size_sample_g),
-						as.integer(b), as.integer(b_star), as.double(Ds), as.double(D), as.double(threshold), as.integer(multi_update), PACKAGE = "BDgraph" )
+						sample_graphs = as.character(sample_graphs), graph_weights = as.double(graph_weights), size_sample_g = as.integer(size_sample_g), counter_all_g = as.integer(counter_all_g),
+						as.integer(b), as.integer(b_star), as.double(Ds), as.double(D), as.integer(multi_update), as.integer(print), PACKAGE = "BDgraph" )
 		}
 
-		if( ( method == "ggm" ) && ( algorithm == "rj_dmh" ) )
+		if( ( method == "ggm" ) && ( algorithm == "rj-dmh" ) )
 		{
 			result = .C( "ggm_DMH_rjmcmc_map", as.integer(iter), as.integer(burnin), G = as.integer(G), as.double(Ts), as.double(Ti), K = as.double(K), as.integer(p), 
 						all_graphs = as.integer(all_graphs), all_weights = as.double(all_weights), K_hat = as.double(K_hat), 
 						sample_graphs = as.character(sample_graphs), graph_weights = as.double(graph_weights), size_sample_g = as.integer(size_sample_g),
-						as.integer(b), as.integer(b_star), as.double(Ds), as.double(D), as.double(threshold), PACKAGE = "BDgraph" )
+						as.integer(b), as.integer(b_star), as.double(Ds), as.double(D), as.integer(print), PACKAGE = "BDgraph" )
 		}
 		
-		if( ( method == "gcgm" ) && ( algorithm == "bd_dmh" ) && ( multi_update == 1 ) )
+		if( ( method == "gcgm" ) && ( algorithm == "bd-dmh" ) && ( multi_update == 1 ) )
 		{
 			result = .C( "gcgm_DMH_bdmcmc_map", as.integer(iter), as.integer(burnin), G = as.integer(G), as.double(Ts), as.double(Ti), K = as.double(K), as.integer(p),
 						as.double(Z), as.integer(R), as.integer(n), as.integer(gcgm_NA),
 						all_graphs = as.integer(all_graphs), all_weights = as.double(all_weights), K_hat = as.double(K_hat), 
 						sample_graphs = as.character(sample_graphs), graph_weights = as.double(graph_weights), size_sample_g = as.integer(size_sample_g),
-						as.integer(b), as.integer(b_star), as.double(D), as.double(Ds), as.double(threshold), PACKAGE = "BDgraph" )
+						as.integer(b), as.integer(b_star), as.double(D), as.double(Ds), as.integer(print), PACKAGE = "BDgraph" )
 		}
  
- 		if( ( method == "gcgm" ) && ( algorithm == "bd_dmh" ) && ( multi_update != 1 ) )
+ 		if( ( method == "gcgm" ) && ( algorithm == "bd-dmh" ) && ( multi_update != 1 ) )
 		{
+			counter_all_g = 0
 			result = .C( "gcgm_DMH_bdmcmc_map_multi_update", as.integer(iter), as.integer(burnin), G = as.integer(G), as.double(Ts), as.double(Ti), K = as.double(K), as.integer(p),
 						as.double(Z), as.integer(R), as.integer(n), as.integer(gcgm_NA),
 						all_graphs = as.integer(all_graphs), all_weights = as.double(all_weights), K_hat = as.double(K_hat), 
-						sample_graphs = as.character(sample_graphs), graph_weights = as.double(graph_weights), size_sample_g = as.integer(size_sample_g),
-						as.integer(b), as.integer(b_star), as.double(D), as.double(Ds), as.double(threshold), as.integer(multi_update), PACKAGE = "BDgraph" )
+						sample_graphs = as.character(sample_graphs), graph_weights = as.double(graph_weights), size_sample_g = as.integer(size_sample_g), counter_all_g = as.integer(counter_all_g),
+						as.integer(b), as.integer(b_star), as.double(D), as.double(Ds), as.integer(multi_update), as.integer(print), PACKAGE = "BDgraph" )
 		}
 
-		if( ( method == "gcgm" ) && ( algorithm == "rj_dmh" ) )
+		if( ( method == "gcgm" ) && ( algorithm == "rj-dmh" ) )
 		{
 			result = .C( "gcgm_DMH_rjmcmc_map", as.integer(iter), as.integer(burnin), G = as.integer(G), as.double(Ts), as.double(Ti), K = as.double(K), as.integer(p),
 						as.double(Z), as.integer(R), as.integer(n), as.integer(gcgm_NA),
 						all_graphs = as.integer(all_graphs), all_weights = as.double(all_weights), K_hat = as.double(K_hat), 
 						sample_graphs = as.character(sample_graphs), graph_weights = as.double(graph_weights), size_sample_g = as.integer(size_sample_g),
-						as.integer(b), as.integer(b_star), as.double(D), as.double(Ds), as.double(threshold), PACKAGE = "BDgraph" )
+						as.integer(b), as.integer(b_star), as.double(D), as.double(Ds), as.integer(print), PACKAGE = "BDgraph" )
 		}	
       
 	}else{
@@ -255,21 +260,21 @@ bdgraph = function( data, n = NULL, method = "ggm", algorithm = "bdmcmc",
 		{
 			result = .C( "ggm_bdmcmc_ma", as.integer(iter), as.integer(burnin), G = as.integer(G), as.double(Ts), K = as.double(K), as.integer(p), 
 						K_hat = as.double(K_hat), p_links = as.double(p_links),
-						as.integer(b), as.integer(b_star), as.double(Ds), as.double(threshold), PACKAGE = "BDgraph" )
+						as.integer(b), as.integer(b_star), as.double(Ds), as.integer(print), PACKAGE = "BDgraph" )
 		}
 		
 		if( ( method == "ggm" ) && ( algorithm == "bdmcmc" ) && ( multi_update != 1 ) )
 		{
 			result = .C( "ggm_bdmcmc_ma_multi_update", as.integer(iter), as.integer(burnin), G = as.integer(G), as.double(Ts), K = as.double(K), as.integer(p), 
 						K_hat = as.double(K_hat), p_links = as.double(p_links),
-						as.integer(b), as.integer(b_star), as.double(Ds), as.double(threshold), as.integer(multi_update), PACKAGE = "BDgraph" )
+						as.integer(b), as.integer(b_star), as.double(Ds), as.integer(multi_update), as.integer(print), PACKAGE = "BDgraph" )
 		}		
     
 		if( ( method == "ggm" ) && ( algorithm == "rjmcmc" ) )
 		{
 			result = .C( "ggm_rjmcmc_ma", as.integer(iter), as.integer(burnin), G = as.integer(G), as.double(Ts), K = as.double(K), as.integer(p), 
 						K_hat = as.double(K_hat), p_links = as.integer(p_links),
-						as.integer(b), as.integer(b_star), as.double(Ds), as.double(threshold), PACKAGE = "BDgraph" )
+						as.integer(b), as.integer(b_star), as.double(Ds), as.integer(print), PACKAGE = "BDgraph" )
 		}
 		
 		if( ( method == "gcgm" ) && ( algorithm == "bdmcmc" ) && ( multi_update == 1 )  )
@@ -277,7 +282,7 @@ bdgraph = function( data, n = NULL, method = "ggm", algorithm = "bdmcmc",
 			result = .C( "gcgm_bdmcmc_ma", as.integer(iter), as.integer(burnin), G = as.integer(G), as.double(Ts), K = as.double(K), as.integer(p),
 						as.double(Z), as.integer(R), as.integer(n), as.integer(gcgm_NA),
 						K_hat = as.double(K_hat), p_links = as.double(p_links),
-						as.integer(b), as.integer(b_star), as.double(D), as.double(Ds), as.double(threshold), PACKAGE = "BDgraph" )
+						as.integer(b), as.integer(b_star), as.double(D), as.double(Ds), as.integer(print), PACKAGE = "BDgraph" )
 		}
     
 		if( ( method == "gcgm" ) && ( algorithm == "bdmcmc" ) && ( multi_update != 1 ) )
@@ -285,7 +290,7 @@ bdgraph = function( data, n = NULL, method = "ggm", algorithm = "bdmcmc",
 			result = .C( "gcgm_bdmcmc_ma_multi_update", as.integer(iter), as.integer(burnin), G = as.integer(G), as.double(Ts), K = as.double(K), as.integer(p),
 						as.double(Z), as.integer(R), as.integer(n), as.integer(gcgm_NA),
 						K_hat = as.double(K_hat), p_links = as.double(p_links),
-						as.integer(b), as.integer(b_star), as.double(D), as.double(Ds), as.double(threshold), as.integer(multi_update), PACKAGE = "BDgraph" )
+						as.integer(b), as.integer(b_star), as.double(D), as.double(Ds), as.integer(multi_update), as.integer(print), PACKAGE = "BDgraph" )
 		}
 
 		if( ( method == "gcgm" ) && ( algorithm == "rjmcmc" ) )
@@ -293,59 +298,59 @@ bdgraph = function( data, n = NULL, method = "ggm", algorithm = "bdmcmc",
 			result = .C( "gcgm_rjmcmc_ma", as.integer(iter), as.integer(burnin), G = as.integer(G), as.double(Ts), K = as.double(K), as.integer(p),
 						as.double(Z), as.integer(R), as.integer(n), as.integer(gcgm_NA),
 						K_hat = as.double(K_hat), p_links = as.integer(p_links),
-						as.integer(b), as.integer(b_star), as.double(D), as.double(Ds), as.double(threshold), PACKAGE = "BDgraph" )
+						as.integer(b), as.integer(b_star), as.double(D), as.double(Ds), as.integer(print), PACKAGE = "BDgraph" )
 		}	
 
 		# for Double Metropolis-Hasting 
-		if( ( method == "ggm" ) && ( algorithm == "bd_dmh" ) && ( multi_update == 1 )  )
+		if( ( method == "ggm" ) && ( algorithm == "bd-dmh" ) && ( multi_update == 1 )  )
 		{
 			result = .C( "ggm_DMH_bdmcmc_ma", as.integer(iter), as.integer(burnin), G = as.integer(G), as.double(Ts), as.double(Ti), K = as.double(K), as.integer(p), 
 						K_hat = as.double(K_hat), p_links = as.double(p_links),
-						as.integer(b), as.integer(b_star), as.double(Ds), as.double(D), as.double(threshold), PACKAGE = "BDgraph" )
+						as.integer(b), as.integer(b_star), as.double(Ds), as.double(D), as.integer(print), PACKAGE = "BDgraph" )
 		}
 
-		if( ( method == "ggm" ) && ( algorithm == "bd_dmh" ) && ( multi_update != 1 ) )
+		if( ( method == "ggm" ) && ( algorithm == "bd-dmh" ) && ( multi_update != 1 ) )
 		{
 			result = .C( "ggm_DMH_bdmcmc_ma_multi_update", as.integer(iter), as.integer(burnin), G = as.integer(G), as.double(Ts), as.double(Ti), K = as.double(K), as.integer(p), 
 						K_hat = as.double(K_hat), p_links = as.double(p_links),
-						as.integer(b), as.integer(b_star), as.double(Ds), as.double(D), as.double(threshold), as.integer(multi_update), PACKAGE = "BDgraph" )
+						as.integer(b), as.integer(b_star), as.double(Ds), as.double(D), as.integer(multi_update), as.integer(print), PACKAGE = "BDgraph" )
 		}		
     
-		if( ( method == "ggm" ) && ( algorithm == "rj_dmh" ) )
+		if( ( method == "ggm" ) && ( algorithm == "rj-dmh" ) )
 		{
 			result = .C( "ggm_DMH_rjmcmc_ma", as.integer(iter), as.integer(burnin), G = as.integer(G), as.double(Ts), as.double(Ti), K = as.double(K), as.integer(p), 
 						K_hat = as.double(K_hat), p_links = as.integer(p_links),
-						as.integer(b), as.integer(b_star), as.double(Ds), as.double(D), as.double(threshold), PACKAGE = "BDgraph" )
+						as.integer(b), as.integer(b_star), as.double(Ds), as.double(D), as.integer(print), PACKAGE = "BDgraph" )
 		}
 		
-		if( ( method == "gcgm" ) && ( algorithm == "bd_dmh" ) && ( multi_update == 1 )  )
+		if( ( method == "gcgm" ) && ( algorithm == "bd-dmh" ) && ( multi_update == 1 )  )
 		{
 			result = .C( "gcgm_DMH_bdmcmc_ma", as.integer(iter), as.integer(burnin), G = as.integer(G), as.double(Ts), as.double(Ti), K = as.double(K), as.integer(p),
 						as.double(Z), as.integer(R), as.integer(n), as.integer(gcgm_NA),
 						K_hat = as.double(K_hat), p_links = as.double(p_links),
-						as.integer(b), as.integer(b_star), as.double(D), as.double(Ds), as.double(threshold), PACKAGE = "BDgraph" )
+						as.integer(b), as.integer(b_star), as.double(D), as.double(Ds), as.integer(print), PACKAGE = "BDgraph" )
 		}
 
-		if( ( method == "gcgm" ) && ( algorithm == "bd_dmh" ) && ( multi_update != 1 ) )
+		if( ( method == "gcgm" ) && ( algorithm == "bd-dmh" ) && ( multi_update != 1 ) )
 		{
 			result = .C( "gcgm_DMH_bdmcmc_ma_multi_update", as.integer(iter), as.integer(burnin), G = as.integer(G), as.double(Ts), as.double(Ti), K = as.double(K), as.integer(p),
 						as.double(Z), as.integer(R), as.integer(n), as.integer(gcgm_NA),
 						K_hat = as.double(K_hat), p_links = as.double(p_links),
-						as.integer(b), as.integer(b_star), as.double(D), as.double(Ds), as.double(threshold), as.integer(multi_update), PACKAGE = "BDgraph" )
+						as.integer(b), as.integer(b_star), as.double(D), as.double(Ds), as.integer(multi_update), as.integer(print), PACKAGE = "BDgraph" )
 		}
 
-		if( ( method == "gcgm" ) && ( algorithm == "rj_dmh" ) )
+		if( ( method == "gcgm" ) && ( algorithm == "rj-dmh" ) )
 		{
 			result = .C( "gcgm_DMH_rjmcmc_ma", as.integer(iter), as.integer(burnin), G = as.integer(G), as.double(Ts), as.double(Ti), K = as.double(K), as.integer(p),
 						as.double(Z), as.integer(R), as.integer(n), as.integer(gcgm_NA),
 						K_hat = as.double(K_hat), p_links = as.integer(p_links),
-						as.integer(b), as.integer(b_star), as.double(D), as.double(Ds), as.double(threshold), PACKAGE = "BDgraph" )
+						as.integer(b), as.integer(b_star), as.double(D), as.double(Ds), as.integer(print), PACKAGE = "BDgraph" )
 		}	
 	}
 
 	K_hat      = matrix( result $ K_hat, p, p ) 
-	last_graph = matrix( result $ G, p, p )
-	last_K     = matrix( result $ K, p, p )
+	last_graph = matrix( result $ G    , p, p )
+	last_K     = matrix( result $ K    , p, p )
 
 	colnames( last_graph ) = colnames( data )
 
@@ -356,16 +361,21 @@ bdgraph = function( data, n = NULL, method = "ggm", algorithm = "bdmcmc",
 		sample_graphs = result $ sample_graphs[ 1 : size_sample_g ]
 		graph_weights = result $ graph_weights[ 1 : size_sample_g ]
 		all_graphs    = result $ all_graphs + 1
-		all_weights   = result $ all_weights	
+		all_weights   = result $ all_weights
+		if( ( algorithm != "rjmcmc" ) & ( multi_update != 1 ) )
+		{ 
+			all_weights = all_weights[ 1 : ( result $ counter_all_g ) ]
+			all_graphs  = all_graphs[  1 : ( result $ counter_all_g ) ] 
+		}
 
 		output = list( sample_graphs = sample_graphs, graph_weights = graph_weights, K_hat = K_hat, 
 					all_graphs = all_graphs, all_weights = all_weights, last_graph = last_graph, last_K = last_K )
 	}else{
 		p_links   = matrix( result $ p_links, p, p ) 
-		if( ( algorithm == "rjmcmc" ) | ( algorithm == "rj_dmh" ) )
+		if( ( algorithm == "rjmcmc" ) | ( algorithm == "rj-dmh" ) )
 		{
 			p_links = p_links / ( iter - burnin )
-			K_hat   = K_hat / ( iter - burnin )
+			K_hat = K_hat / ( iter - burnin )
 		}
 		p_links[ lower.tri( p_links ) ] = 0
 		colnames( p_links ) = colnames( data )
@@ -376,10 +386,11 @@ bdgraph = function( data, n = NULL, method = "ggm", algorithm = "bdmcmc",
 	return( output )   
 }
       
+## ----------------------------------------------------------------------------|
 # summary of bdgraph output
 summary.bdgraph = function( object, vis = TRUE, ... )
 {
-	p_links       = object $ p_links
+	p_links    = object $ p_links
 	p          = nrow( object $ last_graph )
 	dimlab     = colnames( object $ last_graph )
 	selected_g = matrix( 0, p, p, dimnames = list( dimlab, dimlab ) )	
@@ -388,9 +399,9 @@ summary.bdgraph = function( object, vis = TRUE, ... )
 	{
 		sample_graphs = object $ sample_graphs
 		graph_weights = object $ graph_weights
-		max_gWeights = max( graph_weights )
-		sum_gWeights = sum( graph_weights )
-		max_prob_G   = max_gWeights / sum_gWeights
+		max_gWeights  = max( graph_weights )
+		sum_gWeights  = sum( graph_weights )
+		max_prob_G    = max_gWeights / sum_gWeights
 
 		if ( is.null( dimlab ) ) dimlab <- as.character( 1 : p )
 		vec_G    <- c( rep( 0, p * ( p - 1 ) / 2 ) )		
@@ -439,8 +450,8 @@ summary.bdgraph = function( object, vis = TRUE, ... )
 			plot( x = xx, y = weightsg / sum_gWeights, type = "h", main = "Posterior probability of graphs size", ylab = "Pr(graph size|data)", xlab = "Graph size" )
 
 			# plot trace of graph size
-			all_graphs        = object $ all_graphs
-			sizeall_graphs    = sizesample_graphs[ all_graphs ]
+			all_graphs     = object $ all_graphs
+			sizeall_graphs = sizesample_graphs[ all_graphs ]
 			  
 			plot( x = 1 : length( all_graphs ), sizeall_graphs, type = "l", main = "Trace of graph size", ylab = "Graph size", xlab = "Iteration" )
 			
@@ -459,99 +470,130 @@ summary.bdgraph = function( object, vis = TRUE, ... )
 			which_edge       <- which( unlist( strsplit( as.character( sample_graphs[i] ), "" ) ) == 1 )
 			pvec[which_edge] <- pvec[which_edge] + graph_weights[i]
 		}
-		p_links                  <- 0 * selected_g
+		p_links                     <- 0 * selected_g
 		p_links[upper.tri(p_links)] <- pvec / sum_gWeights
 	}
-					  
-	return( list( selected_g = Matrix( selected_g, sparse = TRUE ), p_links = Matrix( p_links, sparse = TRUE ), K_hat = object $ K_hat ) )
+	
+	K_hat = object $ K_hat
+	
+	if( is.null( K_hat ) )			  
+		return( list( selected_g = Matrix( selected_g, sparse = TRUE ), p_links = Matrix( p_links, sparse = TRUE ) ) )
+	else
+		return( list( selected_g = Matrix( selected_g, sparse = TRUE ), p_links = Matrix( p_links, sparse = TRUE ), K_hat = K_hat ) )
 }  
    
+## ----------------------------------------------------------------------------|
 # plot for class bdgraph
-plot.bdgraph = function( x, cut = 0.5, number.g = 1, layout = layout.circle, ... )
+plot.bdgraph = function( x, cut = NULL, number.g = 1, layout = layout.circle, ... )
 {
-	p_links = x $ p_links
-	
-	if( !is.null( x $ graph_weights ) )
+	if( !is.matrix( x ) )
 	{
-		sample_graphs = x $ sample_graphs
-		graph_weights = x $ graph_weights
-		prob_G       = graph_weights / sum( graph_weights )
-		sort_prob_G  = sort( prob_G, decreasing = TRUE )
-
-		p            = nrow( x $ last_graph )
-		dimlab       = colnames( x $ last_graph )
-		list_G       = replicate( number.g, matrix( 0, p, p, dimnames = list( dimlab, dimlab ) ), simplify = FALSE )
-		vec_G        = c( rep( 0, p * ( p - 1 ) / 2 ) )
-
-		if( number.g == 2 ) op <- par( mfrow = c( 1, 2 ), pty = "s" )
-		if( number.g > 2 & number.g < 7 )  op <- par( mfrow = c( 2, number.g %% 2 + trunc( number.g / 2 ) ), pty = "s" )
-
-		for( i in 1 : number.g )
-		{
-			if( number.g > 6 ) dev.new()  
-			indG_i <- sample_graphs[ which( prob_G == sort_prob_G[i] )[1] ]
-			vec_G  <- 0 * vec_G
-			vec_G[ which( unlist( strsplit( as.character(indG_i), "" ) ) == 1 ) ] <- 1
-			list_G[[i]][ upper.tri( list_G[[i]] ) ] <- vec_G
-			G    <- graph.adjacency( list_G[[i]], mode = "undirected", diag = FALSE )
-			main <- ifelse( i == 1, "Graph with highest probability", paste( c( i, "th graph" ), collapse = "" ) )
-			plot.igraph( G, layout = layout, main = main, sub = paste( c( "Posterior probability = ", 
-						round( sort_prob_G[i], 6 ) ), collapse = "" ), ... )	   
-		}
+		if( is.null( x $ graph_weights ) ) cut = 0.5
 		
-		if( number.g > 1 & number.g < 7 ) par( op )
-	}
-	else
-	{
-		if( ( cut < 0 ) || ( cut > 1 ) ) stop( "Value of 'cut' should be between zero and one." )
-		selected_g                = 0 * p_links
-		selected_g[ p_links > cut ]  = 1
-		selected_g[ p_links <= cut ] = 0		
+		if( is.null( cut ) )
+		{
+			sample_graphs = x $ sample_graphs
+			graph_weights = x $ graph_weights
+			prob_G        = graph_weights / sum( graph_weights )
+			sort_prob_G   = sort( prob_G, decreasing = TRUE )
 
-		G    = graph.adjacency( selected_g, mode = "undirected", diag = FALSE )
-		plot.igraph( G, layout = layout, main = "Selected graph", sub = "Edge posterior probability = 0.5", ... )	   		
+			p             = nrow( x $ last_graph )
+			dimlab        = colnames( x $ last_graph )
+			list_G        = replicate( number.g, matrix( 0, p, p, dimnames = list( dimlab, dimlab ) ), simplify = FALSE )
+			vec_G         = c( rep( 0, p * ( p - 1 ) / 2 ) )
+
+			if( number.g == 2 ) op <- par( mfrow = c( 1, 2 ), pty = "s" )
+			if( number.g > 2 & number.g < 7 )  op <- par( mfrow = c( 2, number.g %% 2 + trunc( number.g / 2 ) ), pty = "s" )
+
+			for( i in 1 : number.g )
+			{
+				if( number.g > 6 ) dev.new()  
+				indG_i <- sample_graphs[ which( prob_G == sort_prob_G[i] )[1] ]
+				vec_G  <- 0 * vec_G
+				vec_G[ which( unlist( strsplit( as.character(indG_i), "" ) ) == 1 ) ] <- 1
+				list_G[[i]][ upper.tri( list_G[[i]] ) ] <- vec_G
+				G    <- graph.adjacency( list_G[[i]], mode = "undirected", diag = FALSE )
+				main <- ifelse( i == 1, "Graph with highest probability", paste( c( i, "th graph" ), collapse = "" ) )
+				plot.igraph( G, layout = layout, main = main, sub = paste( c( "Posterior probability = ", 
+							round( sort_prob_G[i], 6 ) ), collapse = "" ), ... )	   
+			}
+			
+			if( number.g > 1 & number.g < 7 ) par( op )
+		}else{
+			if( ( cut < 0 ) || ( cut > 1 ) ) stop( " Value of 'cut' must be between 0 and 1. " )
+
+			p_links = x $ p_links
+			if( is.null( p_links ) ) p_links = plinks( x )
+			selected_g                   = 0 * p_links
+			selected_g[ p_links > cut ]  = 1
+			selected_g[ p_links <= cut ] = 0		
+
+			G = graph.adjacency( selected_g, mode = "undirected", diag = FALSE )
+			plot.igraph( G, layout = layout, main = "Selected graph", sub = paste0( "Edge posterior probability = ", cut ), ... )	   		
+		}
+	}else{
+			G = graph.adjacency( x, mode = "undirected", diag = FALSE )
+			plot.igraph( G, layout = layout, main = "Graph with highest posterior probability", ... )	   		
 	}
 }
      
+## ----------------------------------------------------------------------------|
 # print of the bdgraph output
 print.bdgraph = function( x, round = 3, ... )
 {
-	p_links = x $ p_links
-	
-	if( !is.null( x $ graph_weights ) )
+	if( !is.matrix( x ) )
 	{
-		p             = nrow( x $ last_graph )
-		sample_graphs  = x $ sample_graphs
-		graph_weights  = x $ graph_weights
-		# selected graph
-		max_gWeights  = max( graph_weights )
-		sum_gWeights  = sum( graph_weights )
-		vec_G         = c( rep( 0, p * ( p - 1 ) / 2 ) )
-		indG_max      = sample_graphs[ which( graph_weights == max_gWeights )[1] ]
-		vec_G[ which( unlist( strsplit( as.character( indG_max ), "" ) ) == 1 ) ] = 1
+		p_links = x $ p_links
+		
+		if( !is.null( x $ graph_weights ) )
+		{
+			p             = nrow( x $ last_graph )
+			sample_graphs  = x $ sample_graphs
+			graph_weights  = x $ graph_weights
+			# selected graph
+			max_gWeights  = max( graph_weights )
+			sum_gWeights  = sum( graph_weights )
+			vec_G         = c( rep( 0, p * ( p - 1 ) / 2 ) )
+			indG_max      = sample_graphs[ which( graph_weights == max_gWeights )[1] ]
+			vec_G[ which( unlist( strsplit( as.character( indG_max ), "" ) ) == 1 ) ] = 1
 
-		dimlab     = colnames( x $ last_graph )
-		selected_g = matrix( 0, p, p, dimnames = list( dimlab, dimlab ) )	
-		selected_g[upper.tri(selected_g)] = vec_G
-	}
-	else
-	{
-		selected_g                   = 0 * p_links
-		selected_g[ p_links > 0.5 ]  = 1
-		selected_g[ p_links <= 0.5 ] = 0	
+			dimlab     = colnames( x $ last_graph )
+			selected_g = matrix( 0, p, p, dimnames = list( dimlab, dimlab ) )	
+			selected_g[upper.tri(selected_g)] = vec_G
+		
+		}else{
+			selected_g                   = 0 * p_links
+			selected_g[ p_links > 0.5 ]  = 1
+			selected_g[ p_links <= 0.5 ] = 0	
+		}
+		
+	}else{
+		selected_g = unclass( x )
 	}
 	
 	cat( paste( "" ), fill = TRUE )
 	cat( paste( "Adjacency matrix of selected graph" ), fill = TRUE )
 	cat( paste( "" ), fill = TRUE )
-	printSpMatrix( Matrix( selected_g, sparse = TRUE ), col.names = TRUE, note.dropping.colnames = FALSE )
+	
+	if( !is.matrix( x ) )
+	{	
+		printSpMatrix( Matrix( selected_g, sparse = TRUE ), col.names = TRUE, note.dropping.colnames = FALSE )
+		cat( paste( "" ), fill = TRUE )
+		cat( paste( "Size of selected graph = ", sum( selected_g ) ), fill = TRUE )
+	
+	}else{
+		print( selected_g )
+		cat( paste( "" ), fill = TRUE )
+		cat( paste( "Size of selected graph = ", sum( selected_g ) / 2 ), fill = TRUE )	
+	}
 
-	cat( paste( "" ), fill = TRUE )
-	cat( paste( "Size of selected graph = ", sum( selected_g ) ), fill = TRUE )
-	if( !is.null( x $ graph_weights ) )
-		cat( paste( "Posterior probability of selected graph = ", max_gWeights / sum_gWeights ), fill = TRUE )  
-	else
-		cat( paste( "Edge posterior probability of selected graph = ", 0.5 ), fill = TRUE )
+	if( !is.matrix( x ) )
+	{
+		if( !is.null( x $ graph_weights ) )
+			cat( paste( "Posterior probability of selected graph = ", max_gWeights / sum_gWeights ), fill = TRUE )  
+		else
+			cat( paste( "Edge posterior probability of selected graph = ", 0.5 ), fill = TRUE )
+	}
 	
 	cat( paste( "" ), fill = TRUE )
 } 
