@@ -1,5 +1,19 @@
+// ----------------------------------------------------------------------------|
+//     Copyright (C) 2012-2016 Mohammadi A. and Wit C. E.
+//
+//     This file is part of BDgraph package.
+//
+//     BDgraph is free software: you can redistribute it and/or modify it under 
+//     the terms of the GNU General Public License as published by the Free 
+//     Software Foundation; see <https://cran.r-project.org/web/licenses/GPL-3>.
+//
+//     Maintainer:
+//     Abdolreza Mohammadi: a.mohammadi@rug.nl or a.mohammadi@uvt.nl
+// ----------------------------------------------------------------------------|
+  
 #include "matrix.h"
 
+// ----------------------------------------------------------------------------|
 // Takes square matrix A (p x p) and 
 // retrieves square sub_matrix B (p_sub x p_sub), dictated by vector sub
 void sub_matrix( double A[], double sub_A[], int sub[], int *p_sub, int *p )
@@ -16,6 +30,7 @@ void sub_matrix( double A[], double sub_A[], int sub[], int *p_sub, int *p )
 	}
 }
    
+// ----------------------------------------------------------------------------|
 // Takes symmetric matrix A (p x p) and 
 // retrieves upper part of sub_matrix B (p_sub x p_sub), dictated by vector sub
 void sub_matrix_upper( double A[], double sub_A[], int sub[], int *p_sub, int *p )
@@ -32,6 +47,7 @@ void sub_matrix_upper( double A[], double sub_A[], int sub[], int *p_sub, int *p
 	}
 }
    
+// ----------------------------------------------------------------------------|
 // Takes square matrix A (p x p) and 
 // retrieves vector sub_A which is 'sub' th row of matrix A, minus 'sub' element
 // Likes A[j, -j] in R
@@ -43,6 +59,7 @@ void sub_row_mins( double A[], double sub_A[], int *sub, int *p )
 	memcpy( sub_A + subj, A + subxp + subj + 1, sizeof( double ) * ( pdim - subj - 1 ) );	
 }
    
+// ----------------------------------------------------------------------------|
 // Takes square matrix A (p x p) and 
 // retrieves sub_matrix sub_A(2 x p-2) which is sub rows of matrix A, minus two elements
 // Likes A[(i,j), -(i,j)] in R ONLY FOR SYMMETRIC MATRICES
@@ -70,6 +87,7 @@ void sub_rows_mins( double A[], double sub_A[], int *row, int *col, int *p )
 	}
 }
 
+// ----------------------------------------------------------------------------|
 // Takes square matrix A (p x p) and 
 // retrieves sub_matrix sub_A(p-2 x 2) which is sub cols of matrix A, minus two elements
 // Likes A[-(i,j), (i,j)] in R 
@@ -86,6 +104,7 @@ void sub_cols_mins( double A[], double sub_A[], int *row, int *col, int *p )
 	memcpy( sub_A + p2 + subj - 1, A + subjxp + subj + 1, sizeof( double ) * ( pdim - subj - 1 ) );	
 }
    
+// ----------------------------------------------------------------------------|
 // Takes symmatric matrix A (p x p) and 
 // retrieves A12(1x(p-1)) and A22((p-1)x(p-1))
 // Like A12=A[j, -j], and A22=A[-j, -j] in R
@@ -115,6 +134,7 @@ void sub_matrices1( double A[], double A12[], double A22[], int *sub, int *p )
 	}
 }
     
+// ----------------------------------------------------------------------------|
 // Takes square matrix A (p x p) and 
 // retrieves A11(2x2), A12(2x(p-2)), and A22((p-2)x(p-2))
 // Like A11=A[e, e], A12=A[e, -e], and A22=A[-e, -e] in R
@@ -182,6 +202,7 @@ void sub_matrices( double A[], double A11[], double A12[], double A22[], int *ro
 	}
 }
    
+// ----------------------------------------------------------------------------|
 // Takes square matrix A (p x p) and 
 // retrieves A11_inv(2x2), A21((p-2)x2), and A22((p-2)x(p-2))
 // Like A11_inv=inv( A[e, e] ), A21=A[-e, e], and A22=A[-e, -e] in R
@@ -239,6 +260,7 @@ void sub_matrices_inv( double A[], double A11_inv[], double A21[], double A22[],
 	}
 }
       
+// ----------------------------------------------------------------------------|
 // inverse function for symmetric positive-definite matrices (p x p)
 // WARNING: Matrix you pass is overwritten with the result
 void inverse( double A[], double A_inv[], int *p )
@@ -254,7 +276,8 @@ void inverse( double A[], double A_inv[], int *p )
 	// LAPACK function: computes solution to A * X = B, where A is symmetric positive definite matrix
 	F77_NAME(dposv)( &uplo, &dim, &dim, A, &dim, A_inv, &dim, &info );
 }
-
+    
+// ----------------------------------------------------------------------------|
 // inverse function for symmetric (2 x 2)
 void inverse_2x2( double B[], double B_inv[] )
 {
@@ -265,8 +288,8 @@ void inverse_2x2( double B[], double B_inv[] )
 	B_inv[3]    = B[0] / detB;
 }
     
+// ----------------------------------------------------------------------------|
 // Cholesky decomposition of symmetric positive-definite matrix
-// WARNING: Matrix you pass is overwritten with the result
 // A = U' %*% U
 void cholesky( double A[], double U[], int *p )
 {
@@ -282,9 +305,31 @@ void cholesky( double A[], double U[], int *p )
 			U[j * dim + i] = 0.0;
 }
   
-// To select an edge for BDMCMC algorithm  
-void select_edge( long double rates[], int *index_selected_edge, long double *sum_rates, int *qp )
+// ----------------------------------------------------------------------------|
+// Determinant of symmetric possitive-definite matrix -------------------------|
+// ************  WARNING: Matrix you pass is overwritten **********************
+//  For any symmetric PD Matrix D, we have:
+//                |D| ) = |T| ^ 2
+//  where T is the cholesky decomposition of D. Thus, |T| = \prod_{i = 1}^p T_{ii}
+//  which makes this quite easy.
+void determinant( double A[], double *det_A, int *p )
 {
+	char uplo = 'U';
+	int info, dim = *p, dim1 = dim + 1;
+	
+	F77_NAME(dpotrf)( &uplo, &dim, &A[0], &dim, &info );	
+
+	double result = 1;
+	for( int i = 0; i < dim; i++ ) result *= A[i * dim1];
+	
+	*det_A = result * result;
+}
+    
+// ----------------------------------------------------------------------------|
+// To select an edge for BDMCMC algorithm  
+void select_edge( double rates[], int *index_selected_edge, double *sum_rates, int *qp )
+{
+	//GetRNGstate();
 	int qp_star = *qp;
 
 	// rates = sum_sort_rates
@@ -292,7 +337,7 @@ void select_edge( long double rates[], int *index_selected_edge, long double *su
 		rates[i] += rates[ i - 1 ];
 	
 	*sum_rates   = rates[qp_star - 1];
-	long double random_value = *sum_rates * runif( 0, 1 );
+	double random_value = *sum_rates * runif( 0, 1 );
 
 	// To start, find the subscript of the middle position.
 	int lower_bound = 0;
@@ -308,10 +353,12 @@ void select_edge( long double rates[], int *index_selected_edge, long double *su
 	}
 	
 	*index_selected_edge = ( rates[position] < random_value ) ? ++position : position;
+	//GetRNGstate();
 } 
     
+// ----------------------------------------------------------------------------|
 // To simultaneously select multiple edges for BDMCMC algorithm  
-void select_multi_edges( long double rates[], int index_selected_edges[], int *size_index, long double *sum_rates, int *multi_update, int *qp )
+void select_multi_edges( double rates[], int index_selected_edges[], int *size_index, double *sum_rates, int *multi_update, int *qp )
 {
 	int i, qp_star = *qp, qp_star_1 = qp_star - 1;
 
@@ -319,7 +366,7 @@ void select_multi_edges( long double rates[], int index_selected_edges[], int *s
 	for ( i = 1; i < qp_star; i++ )
 		rates[i] += rates[ i - 1 ];
 	
-	long double max_bound = rates[qp_star_1];
+	double max_bound = rates[qp_star_1];
 	
 	// ---------- for first edge ----------------------------------------------|
 	// To start, find the subscript of the middle position.
@@ -327,7 +374,7 @@ void select_multi_edges( long double rates[], int index_selected_edges[], int *s
 	int upper_bound = qp_star_1;
 	int position    = upper_bound / 2; // ( lower_bound + upper_bound ) / 2;
 
-	long double random_value = max_bound * runif( 0, 1 );
+	double random_value = max_bound * runif( 0, 1 );
 
 	while( upper_bound - lower_bound > 1 )
 	{
@@ -375,6 +422,7 @@ void select_multi_edges( long double rates[], int index_selected_edges[], int *s
 	*sum_rates  = max_bound;
 } 
          
+// ----------------------------------------------------------------------------|
 // computing birth/death rate or alpha for element (i,j)
 // it is for double Metropolis-Hasting algorihtms
 void log_H_ij( double K[], double sigma[], double *log_Hij, int *selected_edge_i, int *selected_edge_j,
@@ -441,7 +489,7 @@ void log_H_ij( double K[], double sigma[], double *log_Hij, int *selected_edge_i
 	*log_Hij = ( log( static_cast<double>(*Dsjj) ) - log( static_cast<double>(a11) ) + *Dsijj * a11 - sum_diag ) / 2;
 }    
 
-// -------------- NEW for Lang codes -------------------------------------------
+// -------------- NEW for Lang codes ------------------------------------------|
 // For Hermitian matrix
 void Hsub_row_mins( double A[], double sub_A[], int *sub, int *p )
 {
@@ -454,6 +502,7 @@ void Hsub_row_mins( double A[], double sub_A[], int *sub, int *p )
 		sub_A[l++] = -A[subxp + i];
 }
       
+// ----------------------------------------------------------------------------|
 // For Hermitian matrix
 void Hsub_rows_mins( double A[], double sub_A[], int *row, int *col, int *p )
 {	
@@ -479,6 +528,7 @@ void Hsub_rows_mins( double A[], double sub_A[], int *row, int *col, int *p )
 	}
 }
        
+// ----------------------------------------------------------------------------|
 // sub_matrices1 for Hermitian matrix
 void Hsub_matrices1( double A[], double A12[], double A22[], int *sub, int *p )
 {
@@ -504,6 +554,7 @@ void Hsub_matrices1( double A[], double A12[], double A22[], int *sub, int *p )
 	}
 }
         
+// ----------------------------------------------------------------------------|
 // sub_matrices for Hermitian matrix
 void Hsub_matrices( double A[], double A11[], double A12[], double A22[], int *row, int *col, int *p )
 {
@@ -553,6 +604,7 @@ void Hsub_matrices( double A[], double A11[], double A12[], double A22[], int *r
 	}
 }
    
+// ----------------------------------------------------------------------------|
 // inverse function for Hermitian (2 x 2)
 void cinverse_2x2( double r_B[], double i_B[], double r_B_inv[], double i_B_inv[] )
 {
@@ -570,9 +622,11 @@ void cinverse_2x2( double r_B[], double i_B[], double r_B_inv[], double i_B_inv[
 	i_B_inv[3] =  ( r_det * i_B[0] - r_B[0] * i_det ) / mod;
 }
 
+// ----------------------------------------------------------------------------|
 // For generating scale-free graphs: matrix G (p x p) is an adjacency matrix
 void scale_free( int *G, int *p )
 {
+	GetRNGstate();
 	int i, j, tmp, total, dim = *p, p0 = 2;
 	double random_value;
 	std::vector<int> size_a( dim ); 
@@ -589,7 +643,6 @@ void scale_free( int *G, int *p )
 	
 	total = 2 * p0;
 	
-	GetRNGstate();
 	for( i = p0; i < dim; i++ )
 	{
 	   random_value = (double) total * runif( 0, 1 );
